/*

   Copyright 2010, 2011 SRI International

   Licensed under the Apache License, Version 2.0 (the "License"); you may
   not use this file except in compliance with the License. You may obtain
   a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   The development was supported by National Science Foundation Grant 0932397:
   A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package lf.app.factorial; 

import cf.api.ApplicationEvent;
import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberNode;
import cf.api.devices.Area;
import lf.api.AppKnowledgeBase;
import lf.api.CPSLogic;
import lf.api.Fact;
import lf.api.Goal;
import alice.tuprolog.Struct;
import alice.tuprolog.Term;

class LogicApp extends CyberApp
{
	Area area; 

	public AppKnowledgeBase akb = new AppKnowledgeBase();

	CPSLogic cpslogic; 

	// params 
	double TheoryGoalSelectionInterval = 0; 

    boolean RandomGoalSelection = true;
	boolean noAlternatives = true; 

	String ForwardRuleFile = "factorial-fr.pl";
	String BackwardRuleFile = "factorial-br.pl";

	// goal string 
	String goal_str = "factorial(15,X)"; 

    public LogicApp(CyberNode node)
    {
		super(node);

		cpslogic = new CPSLogic(node,
                                RandomGoalSelection,
                                noAlternatives,
                                ForwardRuleFile,
                                BackwardRuleFile);
		cpslogic.setFactory(new CPSGoalFactory(), new CPSFactFactory());
	}

	// for performance measure 
	int timeSyncInitDelay = 15;		// sec  
	long startTime = 0;	
	boolean startTimeLogged = false;  

	public void initialize()
	{
		if (name().substring(0,5).equals("node1"))
		{
			postKnowledge(new GoalKnowledge(this, goal_str + ".")); 
		}

        //postEvent(new TheoryGoalSelectionEvent(this, TheoryGoalSelectionInterval));
		ApplicationEvent e = new TheoryGoalSelectionEvent(this, TheoryGoalSelectionInterval);
		e.setAbsTime(timeSyncInitDelay); 
        postEvent(e);
	}

    public void handleEvent(ApplicationEvent e)
    {
		if (!startTimeLogged)
		{
			startTime = System.currentTimeMillis();
			startTimeLogged = true; 
		}

        if(e instanceof TheoryGoalSelectionEvent) handleTheoryGoalSelectionEvent((TheoryGoalSelectionEvent) e);
    }

    public void handleKnowledge(ApplicationKnowledge k)
    {
		if (k instanceof Goal) 
		{
			Goal g = (Goal) k; 

			handleGoalKnowledge(g);
		}
		else if (k instanceof Fact)
		{
			Fact f = (Fact) k;

			handleFactKnowledge(f);

			// currently, robot includes wifi 
			Term t = Term.createTerm(f.term().substring(0,f.term().length()-1));
        	Struct st = (Struct) t;
        	String s = st.getName();
        	if(s.equals("factorial"))
				deliver(f); 
		}
    }

	// currently, node includes wifi
	void deliver(Fact f)
	{
        log(name() + " - network device delivers " + f.term());

        Term t = Term.createTerm(f.term().substring(0,f.term().length()-1));
        Struct st = (Struct) t;
        String s = st.getName();
        if(s.equals("factorial"))
        {
            Term t1 = Term.createTerm(goal_str);

			if (t.match(t1) && name().substring(0,5).equals("node1"))
           	{
                log("Cyber Deliver - " + t.toString() + " " + name() + " at " + time());
                System.out.println("Cyber Deliver - " + t.toString() + " " + name() + " at " + time());

				//TODO 
				System.out.println("<-------------- Mission Accomplished ------------->");
				System.out.println(System.currentTimeMillis() - startTime + " ms");
				System.exit(0);
			}
            else
                log(name() + " - cannot deliver - I am not a destination " + name());
		}
	}

	// application event 
    class TheoryGoalSelectionEvent extends ApplicationEvent
    {
    	public TheoryGoalSelectionEvent(LogicApp app, double delay) { super(app,delay); }
    }

    public void handleTheoryGoalSelectionEvent(TheoryGoalSelectionEvent e)
    {
    	String lf = cpslogic.selectTheoryGoal(akb);
    	if (lf != null)
    	{
        	log(name() + " - solve (by theory goal selection) " + lf);
        	cpslogic.reason(new GoalKnowledge(this, lf));
    	}

    	postEvent(new TheoryGoalSelectionEvent(this, TheoryGoalSelectionInterval));
	}

	// application knowledge 
    void handleFactKnowledge(Fact f)
	{
		log(name() + " - add fact knowledge - " + f.term());

	   	// update akb 
        if (akb.fresh((ApplicationKnowledge) f))
            akb.update((ApplicationKnowledge) f);

        cpslogic.theoryUpdate(akb);
	}

	void handleGoalKnowledge(Goal g)
	{
        log(name() + " - add goal knowledge - " + g.lf());

		// update akb 
        if (akb.fresh((ApplicationKnowledge) g))
            akb.update((ApplicationKnowledge) g);

		cpslogic.theoryUpdate(akb); 
	}
} 
